part of fullauth_api_flutter_client_library;

class AuthService {
  String _authDomain;
  AuthEnvironment _environment;

  AuthService({AuthEnvironment environment, String authDomain}) {
    this._environment = environment;
    this._authDomain = authDomain;
  }

  String getAuthorizeUrl() {
    return ApiHelpers.getAuthUrl(environment: this._environment, domain: this._authDomain);
  }

  String getTokenUrl() {
    return ApiHelpers.getTokenUrl(environment: this._environment, domain: this._authDomain);
  }

  String getTokenInfoUrl() {
    return ApiHelpers.getRevokeTokenUrl(environment: this._environment, domain: this._authDomain);
  }
}

class FullAuthOauthService extends AuthService {
  static FullAuthOAuthServiceBuilder builder({AuthEnvironment environment = AuthEnvironment.live}) {
    return FullAuthOAuthServiceBuilder(environment: environment, authDomain: environment.authDomain);
  }

  String _clientId;
  String _clientSecret;
  Set<String> _scopes;
  String _redirectUri;
  OauthAccessType _accessType;

  FullAuthOauthService(
      {AuthEnvironment environment,
      String authDomain,
      String clientId,
      String clientSecret,
      Set<String> scopes,
      String redirectUri,
      OauthAccessType accessType})
      : super(environment: environment, authDomain: authDomain) {
    this._environment = environment;
    this._clientId = clientId;
    this._clientSecret = clientSecret;
    this._scopes = scopes;
    this._redirectUri = redirectUri;
    this._accessType = accessType;
  }
}

// MARK: Token Generation
extension FullAuthOauthServiceTokenGenerationExtension on FullAuthOauthService {
  Future<OauthAccessToken> requestAccessTokenForGoogleToken(
      {String googleToken, Set<String> scopes, OauthAccessType accessType = OauthAccessType.offline}) async {
    this._validateParams();
    this._validate(googleToken, errorMessage: "invalid google token");
    this._validate(scopes, errorMessage: "invalid scopes, must have at least one scope");
    this._validate(accessType, errorMessage: "invalid accessType, must be either ONLINE or OFFLINE");

    GoogleTokenRequest request = GoogleTokenRequest(clientId: this._clientId, clientSecret: this._clientSecret);
    request.setGoogleToken(googleToken);
    request.setScope(scopes);
    request.setAccessType(accessType);
    return this.makeTokenRequest(request);
  }

  Future<OauthAccessToken> requestAccessTokenForAppleToken(
      {String appleToken, String bundleId, Set<String> scopes, OauthAccessType accessType = OauthAccessType.offline}) async {
    this._validateParams();
    this._validate(appleToken, errorMessage: "invalid apple Token");
    this._validate(bundleId, errorMessage: "invalid app bundleId");
    this._validate(scopes, errorMessage: "invalid scopes, must have at least one scope");
    this._validate(accessType, errorMessage: "invalid accessType, must be either ONLINE or OFFLINE");

    AppleTokenRequest request = AppleTokenRequest(clientId: this._clientId, clientSecret: this._clientSecret);
    request.setAppleToken(appleToken);
    request.setBundleId(bundleId);
    request.setScope(scopes);
    request.setAccessType(accessType);
    return this.makeTokenRequest(request);
  }

  Future<OauthAccessToken> requestAccessTokenForResourceCredentials(
      {String userName, String password, Set<String> scopes, OauthAccessType accessType = OauthAccessType.offline}) async {
    this._validateParams();
    this._validate(userName, errorMessage: "invalid userName");
    this._validate(password, errorMessage: "invalid password");
    this._validate(scopes, errorMessage: "invalid scopes, must have at least one scope");
    this._validate(accessType, errorMessage: "invalid accessType, must be either ONLINE or OFFLINE");

    ResourceOwnerTokenRequest request =
        ResourceOwnerTokenRequest(clientId: this._clientId, clientSecret: this._clientSecret);
    request.setUserName(userName);
    request.setPassword(password);
    request.setScope(scopes);
    request.setAccessType(accessType);
    return this.makeTokenRequest(request);
  }
}

// MARK: Token Actions
extension FullAuthOauthServiceTokenExtension on FullAuthOauthService {
  Future<OauthAccessToken> getTokenInfo({String accessToken}) async {
    this._validate(this._authDomain, errorMessage: "invalid auth domain");
    this._validate(accessToken, errorMessage: "invalid access token");

    var tokenInfoUrl = ApiHelpers.getTokenInfoUrl(environment: this._environment, domain: this._authDomain);
    String accessTokenUrl = "$tokenInfoUrl?access_token=$accessToken";

    try {
      Response response = await http.get(accessTokenUrl);
      this._checkErrorResponse(response); //throws
      OauthAccessToken tokenInfo = this._convertResponseToOauthAccessToken(response); //throws
      tokenInfo.setAccessToken(accessToken);
      return tokenInfo;
    } on TokenResponseException catch (tokenResponseException) {
      throw tokenResponseException; //rethrows
    } on IOException catch (ioException) {
      throw TokenResponseException.errorMessage(ioException.toString()); //rethrows
    }
  }

  Future<bool> revokeAccessToken({String accessToken}) async {
    this._validate(this._authDomain, errorMessage: "invalid auth domain");
    this._validate(accessToken, errorMessage: "invalid access token");

    String revokeTokenUrl = ApiHelpers.getRevokeTokenUrl(environment: this._environment, domain: this._authDomain);
    String url = "$revokeTokenUrl?token=$accessToken";

    try {
      Response response = await http.get(url);
      this._checkErrorResponse(response); //throws
      return true;
    } on TokenResponseException catch (tokenResponseException) {
      throw tokenResponseException; //rethrows
    } on IOException catch (ioException) {
      throw TokenResponseException.errorMessage(ioException.toString()); //rethrows
    }
  }

  Future<OauthAccessToken> refreshAccessToken(
      {String refreshToken, OauthExpiryType expiryType = OauthExpiryType.long}) async {
    this._validateParams();
    this._validate(refreshToken, errorMessage: "invalid refresh token");
    this._validate(expiryType, errorMessage: "invalid expiry type");

    RefreshTokenRequest refreshTokenRequest =
        RefreshTokenRequest(clientId: this._clientId, clientSecret: this._clientSecret);
    refreshTokenRequest.setRefreshToken(refreshToken);
    refreshTokenRequest.setExpiryType(expiryType.type);
    return this.makeTokenRequest(refreshTokenRequest);
  }
}

// MARK: Http
extension FullAuthOauthServiceHttpExtension on FullAuthOauthService {
  Future<OauthAccessToken> makeTokenRequest(OauthTokenRequest tokenRequest) async {
    try {
      String payload = tokenRequest.toFormParams();
      var dict = tokenRequest.getRequestParam();
      print("fullauth request param: $dict");
      var headers = {"Accept": "application/json", "Content-Type": "application/x-www-form-urlencoded"};
      Response response = await http.post(Uri.parse(this.getTokenUrl()), headers: headers, body: payload);
      _checkErrorResponse(response);
      return _convertResponseToOauthAccessToken(response);
    } on TokenResponseException catch (tokenResponseException) {
      throw tokenResponseException;
    } on IOException catch (ioException) {
      throw TokenResponseException.errorMessage(ioException.toString());
    }
  }

  void _checkErrorResponse(Response response) {
    if (response.statusCode != 200) {
      Map<String, dynamic> json = jsonDecode(response.body) as Map<String, dynamic>;
      OauthTokenErrorResponse errorResponse = OauthTokenErrorResponse.fromJson(json);
      if (errorResponse == null) {
        errorResponse = OauthTokenErrorResponse(error: "json_exception", errorDescription: response.reasonPhrase);
      }
      throw TokenResponseException.errorResponse(errorResponse);
    }
  }

  OauthAccessToken _convertResponseToOauthAccessToken(Response response) {
    try {
      Map<String, dynamic> json = jsonDecode(response.body) as Map<String, dynamic>;
      OauthAccessToken token = OauthAccessToken.fromJson(json);
      if (token == null) {
        throw IllegalArgumentException(message: "invalid json content from full auth");
      } else {
        return token;
      }
    } catch (var3) {
      throw TokenResponseException.errorMessage("Got " + response.reasonPhrase);
    }
  }
}

// Utils
extension FullAuthOauthServiceUtilsExtension on FullAuthOauthService {
  void _validate(dynamic value, {String errorMessage}) {
    if (!Helpers.hasValue(value)) {
      throw IllegalArgumentException(message: errorMessage);
    }
  }

  void _validateParams() {
    if (!Helpers.hasValue(this._authDomain)) {
      throw IllegalArgumentException(message: "invalid auth domain");
    } else if (!Helpers.hasValue(this._clientId)) {
      throw IllegalArgumentException(message: "invalid clientId");
    } else if (!Helpers.hasValue(this._clientSecret)) {
      throw IllegalArgumentException(message: "invalid client secret");
    }
  }
}
